const { MongoClient } = require('mongodb')
const fs = require('fs')
const dataloader = require('dataloader')
const _ = require('lodash')
const inflection = require( 'inflection' )
const yves = require('yves')

const graphqlFields = require('graphql-fields')
const { GraphQLScalarType } = require('graphql')
const gql = require('graphql-tag')

const {check, Match} = require('meteor-param-check')
const util = require('./util')
const objectHash = require('object-hash')

const pkg = require('../package.json')
const adapterSchema = require('./adapterSchema')
const adapterMetrics = require('./adapterMetrics')
const adapterIntrospect = require('./adapterIntrospect')
const events = require('events')
const graphQLcolors = require('yonderbox-graphql-colors')
const debugQueryPrefix = `${pkg.name.replace(/-/g,':')}:query`
const debugQuery = yves.debugger(debugQueryPrefix)
const { Kind } = require('graphql/language')

const reportTimeout = 20

const AdapterTypeDefs = []
const AdapterResolvers = []
const AdapterDataSources = {}

const {JSONPath} = require('jsonpath-plus')

const lookupCommand = '_lookup'

const debugPreProcess = yves.debugger(`${pkg.name.replace(/-/g,':')}:preprocess`)
// const appRootPath = require('app-root-path');
// const appPkg = require(path.join(appRootPath.path,'package.json'))

const basicSchema = gql`
"A date"
scalar Date

"A object"
scalar Object

"Root Query for querying data"
#@cacheControl(maxAge: 30)
type Query

"Root Mutation for mutating data"
type Mutation

`
AdapterTypeDefs.push(basicSchema)

const basicResolvers = {
  Date: new GraphQLScalarType({
    name: 'Date',
    description: 'Date custom scalar type',
    parseValue(value) {
      return new Date(value) // value from the client
    },
    serialize(value) {
      return value.toUTCString() // value sent to the client
    },
    parseLiteral(ast) {
      if (ast.kind === Kind.INT) {
        return new Date(ast.value) // ast value is always in string format
      }
      return null
    },
  }),
}

AdapterResolvers.push(basicResolvers)

if (!AdapterDataSources.introspect) {
  AdapterTypeDefs.push(adapterIntrospect.schema)
  AdapterResolvers.push(adapterIntrospect.resolvers)
  AdapterDataSources['introspect']=new adapterIntrospect.datasource({name:pkg.name,version:pkg.version})
}

const { DataSource } = require('apollo-datasource')
class Adapter extends DataSource {
  constructor(options, client, name) {
    check(options, Match.Maybe(Object))
    check(client, Match.Maybe(MongoClient))
    check(name, String)

    super()
    
    this.client = client
    this.schema = null
    this.name = name
    this.pendingOps = 0
    this.debug = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}`)

    this.options = options
    this.debug('options %y',this.options)

    AdapterDataSources[this.name] = this
  }

  static async build(options, name) {
    return new Adapter(options, await Adapter.connectedClient(options), name)
  }

  static async direct(options, name) {
    const mongodb =  await Adapter.build(options, name)
    if (fs.existsSync(options.schema)) {
      mongodb.schema = require(options.schema)
    }
    return mongodb
  }

  static context(args) {
    const ctx = {res: args.res}

    ctx.headers = args.req.headers
    ctx.parameters = (args.req.body && args.req.body.query) ? args.req.body : ( args.req.query.query ? {operationName: args.req.query.operationName, query: args.req.query.query, variables: args.req.query.variables ? JSON.parse(args.req.query.variables) : {}, extensions: args.req.query.extensions ? JSON.parse(args.req.query.extensions) : {} } : null)
    //debug('parameters %y',ctx.parameters)

    ctx.pendingOps = 0
    ctx.start = util.duration()

    ctx.events = new events.EventEmitter()

    ctx.dataloader = !!_.get(ctx.parameters,'variables._dataloader',true)
    return ctx
  }

  operationStart(context) {
    this.pendingOps++
    if (context) {
      context.pendingOps++
      if (context && context.events && context.pendingOps <= 0) {
        debug('send allOperationsStarted %y',this.name)
        context.events.emit('allOperationsStarted') 
      }      
    }
  }

  operationStop(context) {
    this.pendingOps--
    if (context) {
      context.pendingOps--
      if (context && context.events && context.pendingOps <= 0) {
        setTimeout( () => {
          if (context.pendingOps <= 0) {
            debug('send allOperationsStopped %y',this.name)
            context.events.emit('allOperationsStopped') 
          }
        }, reportTimeout)
      }      
    }
  }

  stop(callback) {
    const checkPendingOps = () => {
      if (this.pendingOps>0) {
        this.debug('Pending mongoOps %y',this.pendingOps)
        setTimeout(checkPendingOps,300)
      } else {
        this.debug('Stopping mongodb client')
        this.client.close()
        callback && callback()
      }
    }
    checkPendingOps()
  }

  access(token,context) {
    
    if (_.get(this.options,'authentication.enabled') && _.get(this.options,'authentication.type') === 'token') {
      const filter = {_id:token,active:true, $or: [{expiresAt:null},{expiresAt:{ $gt: new Date()} } ] }
      if (!this.debugAccess) this.debugAccess = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:access`)
      this.debugAccess('acces: %y',filter)
      return this.query('tokens',filter,{limit:1},context).then( data => {
        if (data && data.length) {
          context.token = data[0]
          return {} // otherwise no sub-data in GraphQL
        } else {
          throw new Error('Not Authorized')
        }
      })
    } else {
      return {}
    }
  }

  getDataLoader(collection, filter, options, context) {
    return (ids) => {
      const idFilter = (ids.length == 1) ? { _id: ids[0] } : { _id: { $in: ids} }
      const mrgFilter = this.mergeFilters(idFilter,filter)
      const mrgOptions = Object.assign({},options,{limit:0})
      return this.query(collection,mrgFilter,mrgOptions,context,true).then(data => {
        const itemsByIds = _.keyBy( data, '_id')
        return ids.map(id => (itemsByIds && itemsByIds[id]) || null)
      })
    }
  }

  ensureDataLoader(collection, filter, options, context) {
    if (context) {
      if (filter && !this.sensibleObject(filter)) filter=null
      if (options && !this.sensibleObject(options)) options=null
      const suffix = (filter && options) ? `.filter_${objectHash(filter)}_options_${objectHash(options)}`:
        ( filter ? `.filter_${objectHash(filter)}` : (
          options ? `.options_${objectHash(options)}` : '.prime'
        )
        )
      const dlPath = `adapter.${this.name}.dataloaders.${collection}${suffix}`
      let result = _.get(context,dlPath)
      if (!result) {
        const dlOptions = _.get(this.options,'dataloader.options',{})
        if (dlOptions) {
          if (!dlOptions.maxBatchSize) dlOptions.maxBatchSize=Infinity
        }
        if (!this.debugDataloder) this.debugDataloader = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:dataloader`)
        this.debugDataloader('%y %y %y %y',dlPath,filter,options,dlOptions)
        result = new dataloader(this.getDataLoader(collection, filter, options, context),dlOptions)
        _.set(context,dlPath,result)
      }
      return result
    }
  }

  sensibleObject(filter) {
    return filter && (typeof filter == 'object') && (Object.keys(filter).length > 0)
  }


  mergeFilters(...queries) {
    let result
    const trueQueries = queries.filter( query => this.sensibleObject(query) )
    if (trueQueries.length == 1) {
      result = trueQueries[0]
    } else if (trueQueries.length > 1) {
      result = { $and: trueQueries }
    } else {
      result = {}
    }
    return result
  }

  getSingle(id,collection,filter,options,context) {
    if (id) {
      return this.ensureDataLoader(collection,filter,options,context).load(id)
    } 
    // else {
    //   return this.query(collection,filter,options,context).then( data => ( data.length && data[0] ) )
    // }
  }

  getMany(ids,collection,filter,options,context) {
    //    if (collection.match(/[A-Z]/)) console.trace('camel')
    if (Array.isArray(ids) && ids.length) {
      return this.ensureDataLoader(collection,filter,options,context).loadMany(ids).then( data => (data && (this.sensibleObject(filter) || this.sensibleObject(options))) ? data.filter( r => !!r) : data )
    } 
    // else {
    //   return this.query(collection,filter,options,context)
    // }
  }

  makeQuery(collection, filter, options, context, ignoreDefaults) {
    // if (this.schema) {
    let confine = _.get(this.schema,`collections.${collection}.options.access.token.confine`) || _.get(this.schema,`collections.${collection}.options.access.confine`)
    if (confine) {
      if (!Array.isArray(confine)) confine=[confine]
      if (context.token) {
        /* add confined part to filter based on token */
        const confineQueries = []
        for (let key of confine) {
          const strict = (typeof key == 'object' ? !!key.strict : false)
          let tKey = Array.isArray(key) ? key[1] : (typeof key == 'object' && key.remote ? key.remote : key)
          if (typeof tKey == 'string' && (tKey == 'site' || tKey == 'team' || tKey == 'organisation') && context.token[inflection.pluralize(tKey)] && context.token[inflection.pluralize(tKey)].length ) tKey = inflection.pluralize(tKey)
          if (context.token[tKey]) {
            const cKey = Array.isArray(key) ? key[0] : (typeof key == 'object' && key.local ? key.local : key)
            if (Array.isArray(context.token[tKey])) {
              if (context.token[tKey].length) {
                confineQueries.push( { [cKey] : { $in: strict ? context.token[tKey] : [ /*null,*/ ...context.token[tKey] ] } } )
              }
            } else {
              confineQueries.push( { [cKey] : context.token[tKey] } )
            }
          }
        }
        if (confineQueries.length) {
          filter = this.mergeFilters(filter, /*subquery*/ confineQueries.length == 1 ? confineQueries[0] : { $or: confineQueries })
        }
      }
    }
    filter = Object.assign({},filter)

    if (!ignoreDefaults) {
      options = Object.assign({},options)
      if (!Object.prototype.hasOwnProperty.call(options, 'sort') && _.get(this.schema,`collections.${collection}.options.sort`)) {
        const sort = _.get(this.schema,`collections.${collection}.options.sort`)
        if (sort) options.sort = sort
      }
      if (!Object.prototype.hasOwnProperty.call(options, 'skip') ) {
        const skip = _.get(this.schema,`collections.${collection}.options.skip`,this.options.skip)
        if (skip) options.skip = skip
      }
      if (!Object.prototype.hasOwnProperty.call(options, 'limit') ) {
        const limit = _.get(this.schema,`collections.${collection}.options.limit`,this.options.limit)
        if (limit) options.limit = limit
      }
    }
    // return {collection, filter, options }
    // }
    return {collection, filter, options}
  }

  query(collection, filter, options, context, dataloader) {
    check(collection, String)
    check(filter, Match.Maybe(Object))
    check(options, Match.Maybe({
      skip: Match.Maybe(Number),
      limit: Match.Maybe(Number),
      sort: Match.Maybe(Object),
    }))

    return this.fetch(collection, filter, options, context, dataloader).then( data => {
      if (data && data.length) {
        if (!dataloader) { // If not initiated from dataloader
          const dataLoader = this.ensureDataLoader(collection, /*filter*/null, /*options*/null, context)
          if (dataLoader) {
            data.map (entry => {
              dataLoader.prime(entry._id, entry)
              return entry._id
            })
          }
        }
        if (context && context.res && _.get(this.options,'dependsOn.enabled')) {
          util.dependsOn(context.res,this.options.dependsOn.typePrefix ? `${this.options.root.toLowerCase()}:${inflection.singularize(collection)}` : null,data.map (entry => entry._id) )
        }
      }
      return data
    }).then(data => data.map( entry => {entry.__collection=collection;return entry} ) )
  }

  fetch(collection, pFilter, pOptions, context, dataloader) {
    // if (!dataloader) console.trace('NO dataloader: '+collection)
    const start = util.duration()
    const {filter, options } = this.makeQuery(collection, pFilter, pOptions, context)
    if (filter && options) {
      if (!this.debugQuery) this.debugQuery = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:query`)
      this.debugQuery('%s %y %y %y',dataloader?'dataloader':'direct',collection, filter, options)
      this.operationStart(context)

      let queriesPos
      if (context && context.res) {
        /* reserve the query position by placing a placeholder */
        const queries = _.get(context.res,'introspect.queries',[])
        queries.push('placeholder')
        _.set(context.res,'introspect.queries',queries)
        queriesPos = queries.length-1
      }
      
      if (options && options.limit == 0) delete options.limit
      if (options && options.skip == 0) delete options.skip
    
      return this.client.db(this.options.db)
        .collection(collection)
        .find(filter,Object.assign({},options))
        .toArray()
        .then((data) => {
          if (context && context.res) {
            const duration = util.duration(start)
            function makeReport(report,duration,length) {
              return {
                queries:report.queries + 1 ,
                results:report.results + length ,
                duration:{
                  avg: parseInt((report.duration.sum + duration) / (report.queries + 1)),
                  max: (duration>report.duration.max) ? duration : report.duration.max ,
                  min: (duration<report.duration.min) ? duration : report.duration.min ,
                  sum: report.duration.sum + duration
                },
                via: {
                  dataloader: report.via.dataloader + (dataloader ? 1 : 0),
                  direct: report.via.direct + (!dataloader ? 1 : 0),
                }
              }
            }
            const init = {queries:0,results:0,duration:{min:Number.MAX_SAFE_INTEGER,max:0,avg:0,sum:0},via:{dataloader:0,direct:0}}
            const report =_.get(context.res,`introspect.report.${this.name}.collections.${collection}`,init)
            _.set(context.res,`introspect.report.${this.name}.collections.${collection}`,makeReport(report,duration,data && data.length))
            const report2 =_.get(context.res,`introspect.report.${this.name}.adapter`,init)
            _.set(context.res,`introspect.report.${this.name}.adapter`,makeReport(report2,duration,data && data.length))

            const queries = _.get(context.res,'introspect.queries',[])
            if (queries[queriesPos] == 'placeholder') {
              queries[queriesPos] = {
                adapter: this.name,
                collection,
                filter,
                options,
                via: dataloader?'dataloader':'direct',
                results: data && data.length,
                duration,
              }
              _.set(context.res,'introspect.queries',queries)
            }
          }

          return data
        })/*.catch( e => {
          debug('exception: %y',e)
          reject(e)
        })*/.finally( () => this.operationStop(context) )
    } else {
      throw new Error('No filter or options provided for '+this.name)
    }
  }

  explain(collection, pFilter, pOptions, context) {
    const {filter, options} = this.makeQuery(collection, pFilter, pOptions, context)
    if (filter && options) {
      if (!this.debugExplain) this.debugExplain = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:explain`)
      this.debugExplain('%y %y %y',collection, filter, options)
      this.operationStart(context)
      return this.client.db(this.options.db)
        .collection(collection)
        .find(filter,Object.assign({},options))
        .explain()
        .finally( () => this.operationStop(context) )
    } else {
      throw new Error('No filter or options provided for '+this.name)
    }
  }

  update(collection, pFilter, pModifier, pOptions, context) {
    const {filter, options } = this.makeQuery(collection, pFilter, pOptions, context, true)
    if (filter && options) {
      if (!this.debugUpdate) this.debugUpdate = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:update`)
      this.debugUpdate('%y %y %y %y',collection, filter, pModifier,options)
      this.operationStart(context)
      return this.client.db(this.options.db)
        .collection(collection)
        .findOneAndUpdate(filter, pModifier, Object.assign({},options))
        .catch( err => {
//          console.log(err)
          throw err
        })
        .finally( () => this.operationStop(context) )
    } else {
      throw new Error('No filter or options provided for '+this.name)
    }
  }

  insert(collection, pFilter, pOptions, context) {
    const {filter, options } = this.makeQuery(collection, pFilter, pOptions, context, true)
    if (filter && options) {
      if (!this.debugInsert) this.debugInsert = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:insert`)
      this.debugInsert('%y %y',collection, pFilter)
      this.operationStart(context)
      return this.client.db(this.options.db)
        .collection(collection)
        .insertOne(pFilter)
        .catch( err => {
//          console.log(err)
          throw err
        })
        .finally( () => this.operationStop(context) )
    } else {
      throw new Error('No filter or options provided for '+this.name)
    }
  }

  delete(collection, pFilter, pOptions, context) {
    const {filter, options } = this.makeQuery(collection, pFilter, pOptions, context, true)
    if (filter && options) {
      if (!this.debugInsert) this.debugInsert = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:insert`)
      this.debugInsert('%y %y',collection, pFilter)
      this.operationStart(context)
      return this.client.db(this.options.db)
        .collection(collection)
        .deleteOne(pFilter)
        .catch( err => {
//          console.log(err) 
          throw err
        })
        .finally( () => this.operationStop(context) )
    } else {
      throw new Error('No filter or options provided for '+this.name)
    }
  }

  count(collection, pFilter, pOptions, context) {
    const {filter, options } = this.makeQuery(collection, pFilter, pOptions, context)
    if (filter && options) {
      if (!this.debugCount) this.debugCount = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:count`)
      this.debugCount('%y %y %y',collection, filter, options)
      this.operationStart(context)
      return this.client.db(this.options.db)
        .collection(collection)
        .find(filter,Object.assign({},options))
        .count()
        .finally( () => this.operationStop(context) )
    } else {
      throw new Error('No filter or options provided for '+this.name)
    }
  }

  createCollection(collection, options) {
    this.operationStart()
    return this.client.db(this.options.db)
      .createCollection(collection, options)
      .finally( () => this.operationStop() )
  }

  dropCollection(collection, options) {
    this.operationStart()
    return this.client.db(this.options.db)
      .dropCollection(collection, options)
      .finally( () => this.operationStop() )
  }

  createIndex(collection, specs, options) {
    this.operationStart()
    if (!this.debugCreateIndex) this.debugCreateIndex = yves.debugger(`${pkg.name.replace(/-/g,':')}:${this.name}:createIndex`)
    this.debugCreateIndex('%y %y %y',collection, specs, options)
    return this.client.db(this.options.db)
      .collection(collection)
      .createIndexes([{key:specs, ...options}])
      .catch( err => {
        if (err.code == 11000 || err.codeName == 'DuplicateKey') {
          console.error(`Fallback to non-unique index due to error: ${err.toString()}`)
          /* Fall back to non-unique index */
          delete(options.unique)
          this.operationStart()
          return this.client.db(this.options.db)
            .collection(collection)
            .createIndex(specs, options)
            .finally( () => this.operationStop() )
        } else {
          console.error(err)
        }
      })
      .finally( () => this.operationStop() )
  }

  async introspect(collection, filter, options, context, info) {
    check(collection, String)
    check(filter, Match.Maybe(Object))
    check(options, Match.Maybe(Object))

    const fieldInfo = graphqlFields(info)

    const all = [
      fieldInfo.collection || fieldInfo.filter || fieldInfo.options ? this.makeQuery(collection,filter,options,context) : null,
      fieldInfo.count ? this.count(collection,filter,options,context) : null,
      fieldInfo.total ? this.count(collection,filter,{...options, ...{skip:0,limit:0}},context) : null,
      fieldInfo.data ? this.fetch(collection,filter,options,context) : null,
      fieldInfo.explain ? this.explain(collection,filter,options,context): null,
    ]

    this.operationStart(context)
    return Promise.all(all).then( data => {
      if (data) {
        let result = {}
        if (fieldInfo.collection || fieldInfo.filter || fieldInfo.options) result = Object.assign(result, data[0])
        result.collection = collection
        if (fieldInfo.count) result.count = data[1]
        if (fieldInfo.total) result.total = data[2]
        if (fieldInfo.data) result.data = data[3]
        if (fieldInfo.explain) result.explain = data[4]
        if (fieldInfo.skip) result.skip = options && options.skip
        if (fieldInfo.limit) result.limit = options && options.limit
        if (fieldInfo.sort) result.sort = options && options.sort
        // this.debug('introspect %y',result)
        return result
      }
    }).finally( () => this.operationStop(context) )
  }

  async introspects(collections, filter, options, context, info) {
    check(collections, [String])
    check(filter, Match.Maybe(Object))
    check(options, Match.Maybe(Object))

    const all = collections.map( collection => this.introspect(collection, filter, options, context, info) )

    this.operationStart(context)
    return Promise.all(all).then( data => data.sort((a, b) => a.collection.localeCompare(b.collection)) ).finally( () => this.operationStop(context) )
  }

  static moduleSetup(options,name) {
    return adapterSchema.moduleSetup(options,name)
  }


  static async connectedClient(options) {
    const mongoClient = new MongoClient(options.url,Object.assign({ useNewUrlParser: true, useUnifiedTopology: true},options.options))
    return await mongoClient.connect().then( d => {
      debug('MongoDB connected')
      return d
    }).catch( e => {
      debug('MongoDB connect exceptions: %y',e)
    })
  }


  static preProcessVariables(req, res) {
    return new Promise( async (resolve, reject) => {
      const preprocessedCache = {}
      async function preprocessVariablesRecursive(param) {
        if (typeof(param) == 'object') {
          for (let key in param) {
            param[key] = await preprocessVariablesRecursive(param[key])
          }
          if (param && param[lookupCommand]) {
            if (typeof(param[lookupCommand]) == 'object' && typeof(param[lookupCommand]._adapter) == 'string' && typeof(param[lookupCommand]._collection) == 'string') {
              if (AdapterDataSources[param[lookupCommand]._adapter] && typeof(AdapterDataSources[param[lookupCommand]._adapter].query) == 'function') {
                
                const paramHash = objectHash(param)
                debugPreProcess('execute %y',param)
                if (!preprocessedCache[paramHash]) {
                  const filter = adapterSchema.makeFilterWithArgs(param[lookupCommand],param[lookupCommand]._collection,AdapterDataSources[param[lookupCommand]._adapter].schema)
                  const options = adapterSchema.makeOptionsWithArgs(param[lookupCommand])
                  const context = { res }
                  const data = await AdapterDataSources[param[lookupCommand]._adapter].query(param[lookupCommand]._collection, filter, options, context)
                  preprocessedCache[paramHash] = data.map(entry => entry[param[lookupCommand]._field || '_id'])
                }
                return preprocessedCache[paramHash]
              } else {
                throw new Error(`Unknown or incompattible adapter '${param[lookupCommand]._adapter}' for ${lookupCommand} command variable preprocessing`)
                // debugPreProcess('error %y',new Error(`Unknown or incompatible adapter '${param[lookupCommand]._adapter}' for variable preprocessing with ${lookupCommand}`))
              }
            } else {
              throw new Error(`Badly formed '${lookupCommand}' command for variable preprocessing`)
            }
          }
        }
        return param
      }
    
      if (req && (req.query && req.query.variables) || (req.body && req.body.variables)) {
        const container =  (req.body && req.body.variables) ? 'body' : 'query'
    
        debugPreProcess('original variables (in %s): %y',container,req.body.variables)
        try {
          req[container].variables = await preprocessVariablesRecursive(req[container].variables)
          debugPreProcess('processed variables (in %s): %y',container,req.body.variables)
          resolve()
        } catch(e) {
          debugPreProcess('error %y',e)
          reject(e)
        }
      } else {
        resolve()
      }
    })
  }

  static register(name, setup) {
    if (!AdapterDataSources[name]) {
      const adapter = Adapter.moduleSetup(setup,name)
      if (adapter) {
        if (adapter.schema) AdapterTypeDefs.push(adapter.schema)
        if (adapter.resolvers) AdapterResolvers.push(adapter.resolvers)
        Adapter.build(setup,name)
      }
    }
  }

  static typeDefs() {
    return AdapterTypeDefs
  }

  static resolvers() {
    debug('resolvers %y %y',this.name,AdapterResolvers)
    return AdapterResolvers
  }

  static dataSources() {
    return AdapterDataSources
  }

  static schemaDirectives() {
    return {}
  }

  static plugins({useJsonPath}) {
    return[
      {
        requestDidStart({ request }) {
          // debug('request %y',request);
          return {
            willSendResponse({ response }) {
              // debug('responbse %y',response);
              if (useJsonPath) {
                const _path = _.get(request,'variables._path')
                if (_path) {
                  // https://goessner.net/articles/JsonPath/index.html#e2
                  response.data = JSONPath({path: _path, preventEval: true, json:response.data })
                }
              }
            },
          }
        },
      },
    ]
  }

  static applyMiddleware({app,redirectRoot,preProcessVariables,logGraphQL,metrics, monitor}) {
    if (app) {
      if (metrics) {
        adapterMetrics.instrument(app)
      }
      if (redirectRoot) {
        app.get('/', function(req, res) {
          res.redirect(redirectRoot)
        })        
      }
      if (monitor) {
        app.get('/monitor', function(req, res) {
          res.send('OK')
        })        
      }
      if (preProcessVariables) {
        app.use(async function(req, res, next) {
          Adapter.preProcessVariables(req,res)
            .then( () => next() )
            .catch( e => {
              res.status(400).send({
                message: e.toString()
              })
            })
        })
      }
      if (logGraphQL) {
        app.use(function (req, res, next) {
          if (yves.debugger().enabled(debugQueryPrefix)) {
            let args
            try {
              args = (req.body && req.body.query) ? req.body : ( req.query.query ? {operationName: req.query.operationName, query: req.query.query, variables: req.query.variables ? JSON.parse(req.query.variables) : {}, extensions: req.query.extensions ? JSON.parse(req.query.extensions) : {} } : null)
            } catch (e) {
              console.error(e)
            }
      
            if (args && !args.silentOperation) {
              try {
                const q = graphQLcolors(args)
                if (q) {
                  debugQuery(q)
                } else {
                  debugQuery('no color: %y %y', req.body, req.query)
                }
              } catch(e) {
                console.error(e)
              }
            }
          }
          next()
        })
      }
    }    
  }

  static setIntrospect({name, version, setup}) {
    AdapterDataSources.introspect.set({name, version, setup})
  }
}

module.exports = Adapter
